/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */





#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "liberty.h"

static void analyze_semeai(int m, int n, int i, int j, int color);


/* semeai() searches for pairs of adjacent dead worms. If such a pair
 * is found, the number of liberties is compared, counting mutual
 * liberties only for strings of genus one. If the number of 
 * liberties is equal, it looks for an external liberty of the
 * opponent's string to fill. The parameter (ai, aj) points to
 * the origins of a pair of adjacent dead worms already evaluated,
 * to avoid redundancy.
 *
 * FIXME:
 *
 * The present implementation of the semeai code sometimes 
 * returns wrong answers. In reality, semeai can only be 
 * analyzed through reading. It will be awhile before GNU
 * Go is able to do this accurately. In the meantime, a
 * good idea would be to return a lower weight (say 60) 
 * if the situation differs from criticality by just
 * one liberty and the opponent has an outside liberty
 * which can be filled.
 */

void
semeai(int color)
{
  int d1, d2;
  int k;
  int ai = -1, aj = -1;
  int bi = -1, bj = -1;
  int other = OTHER_COLOR(color);

  TRACE("Semeai Player is THINKING for %s!\n", 
	color_to_string(color));

  for (d1=0; d1<number_of_dragons; d1++) {
    if (DRAGON(d1).color != color
	|| (dragon2[d1].safety != DEAD
	    && dragon2[d1].safety != TACTICALLY_DEAD))
      continue;

    for (k=0; k<dragon2[d1].neighbors; k++) {
      d2 = dragon2[d1].adjacent[k];
      if (DRAGON(d2).color != other
	  || (dragon2[d2].safety != DEAD
	      && dragon2[d2].safety != TACTICALLY_DEAD))
	continue;

      /* Dragons d1 (our) and d2 (opponent) are adjacent and both DEAD
       * or TACTICALLY_DEAD. Analyze the semeai.
       */
      ai = DRAGON(d1).origini;
      aj = DRAGON(d1).originj;
      bi = DRAGON(d2).origini;
      bj = DRAGON(d2).originj;
      analyze_semeai(ai, aj, bi, bj, color);
    }
  }
}


/* liberty_of_dragon(i, j, m, n) returns true if the vertex at (i, j) is a
 * liberty of the dragon with origin at (m, n).
 */

static int 
liberty_of_dragon(int i, int j, int m, int n)
{
  if (i > 0
      && dragon[i-1][j].origini == m
      && dragon[i-1][j].originj == n)
    return 1;

  if (i < board_size - 1
      && dragon[i+1][j].origini == m
      && dragon[i+1][j].originj == n)
    return 1;

  if (j > 0
      && dragon[i][j-1].origini == m
      && dragon[i][j-1].originj == n)
    return 1;

  if (j < board_size - 1
      && dragon[i][j+1].origini == m
      && dragon[i][j+1].originj == n)
    return 1;

  return 0;
}


/* 
 *
 * Rules for playing semeai (capturing races). Let M be the number of
 * liberties of my group, excluding common liberties; let Y be the
 * number of liberties of your group, excluding common liberties;
 * and let C be the number of common liberties. 
 * 
 *             If both groups have zero eyes:
 * 
 * (1) If C=0 and M=Y, whoever moves first wins. CRITICAL.
 * 
 * (2) If C=0 and M>Y, I win.
 * 
 * (3) If C=0 and M<Y, you win.
 * 
 * (4) If C>0 and M >= Y+C then your group is dead and mine is alive.
 * 
 * (5) If C>0 and M = Y+C-1 then the situation is CRITICAL. If M=0,
 * then Y=0 and C=1. In this case, whoever moves first kills.
 * If M>0, then I can kill or you can make seki.
 * 
 * (6) If M < Y+C-1 and Y < M+C-1 then the situation is seki.
 * 
 * (7) If C>0 and Y=M+C-1 then the situation is CRITICAL. If Y=0,
 * then M=0 and C=1 as in (5). If Y>0, then you can kill or I can
 * make seki.
 * 
 * (8) If C>0 and Y > M+C then your group is alive and mine is dead.
 * 
 *              If both groups have one eye:
 * 
 * In this case M > 0 and Y > 0.
 * 
 * (1) If M>C+Y then I win.
 * 
 * (2) If Y>C+M then you win.
 * 
 * (3) If C=0 and M=Y then whoever moves first kills. CRITICAL.
 * 
 * (4) If C>0 and M=C+Y then I can kill, you can make seki. CRITICAL.
 * 
 * (5) If C>0 and M<C+Y, Y<C+M, then the situation is seki. 
 * 
 * (6) If C>0 and Y=C+M, then you can kill, I can make seki. CRITICAL.
 * 
 *            If I have an eye and you dont:
 * 
 * In this case, M > 0. This situation (me ari me nashi) can
 * never be seki. The common liberties must be filled by you,
 * making it difficult to win.
 * 
 * 
 * (1) If M+C>Y then I win.
 * 
 * (2) If M+C=Y then whoever moves first wins. CRITICAL.
 * 
 * (3) If M+C<Y then you win.
 * 
 *            If you have an eye and I don't
 * 
 * In this case, Y > 0. 
 * 
 * (1) If Y+C>M you win.
 * 
 * (2) If Y+C=M whoever moves first wins. CRITICAL.
 * 
 * (3) If Y+C<M I win.
 * 
 */


static void
analyze_semeai(int ai, int aj, int bi, int bj, int color)
{
  int mylibs = 0, yourlibs = 0, commonlibs = 0;
  int yourlibi = -1, yourlibj = -1;
  int commonlibi = -1, commonlibj = -1;
  int i, j;
  int bestlib = 0;
  int commonlib_value = 0;
  int our_status = UNKNOWN;
  int your_status = UNKNOWN;
  int ti, tj;
  
  DEBUG(DEBUG_SEMEAI, "semeai_analyzer : %m (me) vs %m (them)\n",
	ai, aj, bi, bj);

  /* FIXME : rather than looking at all the empty squares on the
   * board, can we not just countlib(i,j) and get a list of
   * the liberties of group at (i,j), and then look at each of
   * those to see which are also liberties of (m,n)
   */

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (same_dragon(i, j, ai, aj) || same_dragon(i, j, bi, bj))
	dragon[i][j].semeai = 1;

      if (p[i][j] != EMPTY)
	continue;
      
      if (liberty_of_dragon(i, j, ai, aj)) {
	if (liberty_of_dragon(i, j, bi, bj)) {
	  DEBUG(DEBUG_SEMEAI, "- %m is a common liberty\n", i, j);
	  commonlibs++;
	  /* Examine whether it's useful to play at this common liberty. */
	  if (trymove(i, j, color, "semeai", -1, -1)) {
	    if (!attack(ai, aj, NULL, NULL)) {
	      int delta;
	      int blib;
	      /* The string at (bi, bj) may have been captured by
	       * the (i, j) move.
	       */
	      if (!p[bi][bj])
		blib = 0;
	      else
		blib = countlib(bi, bj);
	      
	      delta = ((countlib(ai, aj) - blib)
		       - (worm[ai][aj].liberties - worm[bi][bj].liberties));
	      if (commonlibi == -1 || delta > commonlib_value) {
		commonlibi = i;
		commonlibj = j;
		commonlib_value = delta;
	      }
	    }
	    popgo();
	  }
	}
	else {
	  DEBUG(DEBUG_SEMEAI, "- %m is a liberty of %m only\n", 
		i, j, ai, aj);
	  mylibs++;
	}
      }
      else {
	if (liberty_of_dragon(i, j, bi, bj)) {
	  DEBUG(DEBUG_SEMEAI, "- %m is a liberty of %m only\n",
		i, j, bi, bj);
	  yourlibs++;
	  if (trymove(i, j, color, "semeai", -1, -1)) {
	    if (!attack(ai, aj, NULL, NULL)) {
	      int currentlib = countlib(i, j);
	      if (yourlibi == -1 || currentlib > bestlib) {
		yourlibi = i;
		yourlibj = j;
		bestlib = currentlib;
	      }
	    }
	    popgo();
	  }
	}
      }
    }

  /* Set (ti, tj) to a supposedly relevant move in the semeai. */
  if (commonlib_value > 0) {
    ti = commonlibi;
    tj = commonlibj;
  }
  else if (yourlibi != -1) {
    /* I can take off this liberty */
    ti = yourlibi;
    tj = yourlibj;
  }
  else if (commonlibi != -1) {
    ti = commonlibi;
    tj = commonlibj;
  }
  else {
    ti = -1;
    tj = -1;
  }

  DEBUG(DEBUG_SEMEAI, ": considering move further %m\n", ti, tj);

  /* To avoid blunders we verify that the move under consideration is
   * tactically safe. As the semeai player doesn't know about
   * squeezing throw ins yet, this requirement is very reasonable.
   *
   * FIXME: It looks like this condition would always be satisfied
   * after having passed the attack() calls above.
   */
  if (ti != -1 && !safe_move(ti, tj, color)) {
    DEBUG(DEBUG_SEMEAI, "Move at %m not tactically safe.\n", ti, tj);
    ti = -1;
    tj = -1;
  }

  /* Neither dragon has an eye. */
  if (dragon[ai][aj].genus == 0 && dragon[bi][bj].genus == 0) {
    int delta = mylibs + commonlibs - yourlibs;
    if (commonlibs == 0)
      delta++;

    if (delta > 1)
      our_status = ALIVE;
    if (delta == 1)
      our_status = CRITICAL;
    if (delta < 1)
      our_status = DEAD;

    delta = yourlibs + commonlibs - mylibs;
    if (commonlibs == 0)
      delta++;

    if (delta > 1)
      your_status = ALIVE;
    if (delta == 1)
      your_status = CRITICAL;
    if (delta < 1)
      your_status = DEAD;
  }

  /* One eye each. */
  if (dragon[ai][aj].genus == 1 && dragon[bi][bj].genus == 1) {
    int delta = mylibs + commonlibs - yourlibs;
    if (commonlibs == 0)
      delta++;

    if (delta > 0)
      our_status = ALIVE;
    if (delta == 0)
      our_status = CRITICAL;
    if (delta < 0)
      our_status = DEAD;

    delta = yourlibs + commonlibs - mylibs;
    if (commonlibs == 0)
      delta++;

    if (delta > 0)
      your_status = ALIVE;
    if (delta == 0)
      your_status = CRITICAL;
    if (delta < 0)
      your_status = DEAD;
  }

  /* We have an eye, you don't. */
  if (dragon[ai][aj].genus == 1 && dragon[bi][bj].genus == 0) {
    int delta = mylibs + commonlibs - yourlibs;
    if (delta > 0) {
      our_status = ALIVE;
      your_status = DEAD;
    }
    if (delta == 0) {
      our_status = CRITICAL;
      your_status = CRITICAL;
    }
    if (delta < 0) {
      our_status = DEAD;
      your_status = ALIVE;
    }
  }

  /* You have an eye, we don't. */
  if (dragon[ai][aj].genus == 0 && dragon[bi][bj].genus == 1) {
    int delta = yourlibs + commonlibs - mylibs;
    if (delta > 0) {
      your_status = ALIVE;
      our_status = DEAD;
    }
    if (delta == 0) {
      your_status = CRITICAL;
      our_status = CRITICAL;
    }
    if (delta < 0) {
      your_status = DEAD;
      our_status = ALIVE;
    }
  }

  /* We would like to assert this, but until the detection of
   * inessential dragons has been improved it doesn't look like we can
   * guarantee anything. A typical problem is that the basic genus
   * computation may come up with two eyes but the mouse code more
   * accurately says it's in fact dead. If the eyespace includes
   * opponent stones, which have not been classified as inessential,
   * the resulting semeai will seem to be of the type "2 eyes against
   * 0" which we for natural reasons won't consider.
   */
#if 0
  assert(our_status != UNKNOWN);
  assert(your_status != UNKNOWN);
#else
  if (our_status == UNKNOWN || your_status == UNKNOWN) {
    TRACE("Semeai analysis failed, giving up.\n");
    return;
  }
#endif
  
  /* FIXME: We generate mouse_attack/mouse_defense move reasons below.
   *        This works like we want but is kind of improper and
   *        potentially confusing.
   */
  
  if (our_status == ALIVE) {
    DEBUG(DEBUG_SEMEAI, "We are alive.\n");
    DRAGON2(ai, aj).safety = ALIVE_IN_SEKI;
    dragon[ai][aj].status = ALIVE;
  }
  else if (our_status == CRITICAL) {
    if (ti == -1)
      DEBUG(DEBUG_SEMEAI, "Trouble. Our dragon unstable but no move.\n");
    else {
      DEBUG(DEBUG_SEMEAI, "%m is critical if we are to live\n", ti, tj);
      DRAGON2(ai, aj).safety = CRITICAL;
      dragon[ai][aj].status = CRITICAL;
      TRACE("semeai: added mouse defense of %m at %m.\n", ai, aj, ti, tj);
      add_mouse_defense_move(ti, tj, ai, aj);
    }
  }
  else {
    DEBUG(DEBUG_SEMEAI, "We are dead\n");
  }
  
  if (your_status == ALIVE) {
    DEBUG(DEBUG_SEMEAI, "You are alive.\n");
    DRAGON2(bi, bj).safety = ALIVE_IN_SEKI;
    dragon[bi][bj].status = ALIVE;
  }
  else if (your_status == CRITICAL) {
    if (ti == -1) {
      DEBUG(DEBUG_SEMEAI, "Trouble. Your dragon unstable but no move.\n");
      DRAGON2(bi, bj).safety = ALIVE_IN_SEKI;
      dragon[bi][bj].status = ALIVE;
    }
    else {
      DEBUG(DEBUG_SEMEAI, "%m is critical if we are to kill\n", ti, tj);
      DRAGON2(bi, bj).safety = CRITICAL;
      dragon[bi][bj].status = CRITICAL;
      TRACE("semeai: added mouse attack of %m at %m.\n", bi, bj, ti, tj);
      add_mouse_attack_move(ti, tj, bi, bj);
    }
  }
  else {
    DEBUG(DEBUG_SEMEAI, "You are dead\n");
  }
}

/* revise_semeai(color) changes the status of any DEAD dragon of
 * OPPOSITE_COLOR(color) which occurs in a semeai to UNKNOWN.
 * It returns true if such a dragon is found.
 */

int
revise_semeai(int color)
{
  int m, n;
  int found_one = 0;
  int other = OTHER_COLOR(color);

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((dragon[m][n].semeai) 
	  && (dragon[m][n].matcher_status == DEAD) 
	  && (dragon[m][n].color == other)) 
      {
	found_one = 1;
	dragon[m][n].matcher_status = UNKNOWN;
	if (dragon[m][n].origini == m && dragon[m][n].originj == n)
	  TRACE("revise_semeai: changed status of dragon %m from DEAD to UNKNOWN\n",
		m, n);
      }
    }

  return found_one;
}


/* small_semeai() addresses a deficiency in the reading code:
 * for reasons of speed, savestone3 and savestone4 do not
 * sufficiently check whether there may be an adjoining string
 * which can be attacked. So they may overlook a defensive
 * move which consists of attacking an adjoining string.
 *
 * small_semeai(), called by make_worms() searches for a 
 * string A with 3 or 4 liberties such that worm[A].attack_code != 0.
 * If there is a string B next to A (of the opposite color)
 * such that worm[B].attack_code != 0, the following action is
 * taken: if worm[A].liberties == worm[B].liberties, then
 * worm[A].defend is set to worm[B].defend and vice versa;
 * and if worm[A].liberties > worm[B].liberties, then worm[A].defendi
 * is set to -1.
 */

void
small_semeai()
{
  int i, j;
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if (p[i][j]
	  && ((worm[i][j].liberties == 3) || (worm[i][j].liberties == 4))
	  && (worm[i][j].attack_code != 0)) {
	if ((i>0) && (p[i-1][j] == OTHER_COLOR(p[i][j])))
	  small_semeai_analyzer(i, j, i-1, j);
	if ((i<board_size) && (p[i+1][j] == OTHER_COLOR(p[i][j])))
	  small_semeai_analyzer(i, j, i+1, j);
	if ((j>0) && (p[i][j-1] == OTHER_COLOR(p[i][j])))
	  small_semeai_analyzer(i, j, i, j-1);
	if ((j<board_size) && (p[i][j+1] == OTHER_COLOR(p[i][j])))
	  small_semeai_analyzer(i, j, i, j+1);
      }
}

/* Helper function for small_semeai. Tries to resolve the
 * semeai between (i,j) and (m,n), possibly revising points
 * of attack and defense.
 */

void
small_semeai_analyzer(int i, int j, int m, int n)
{
  int ai, aj;
  int color=p[i][j];
  int other=p[m][n];

  if ((worm[m][n].attack_code == 0) || (worm[m][n].liberties < 3))
    return;
  if ((worm[i][j].attack_code == 0) || (worm[i][j].liberties < 3))
    return;


  /* FIXME: There are many more possibilities to consider */
  if (trymove(worm[i][j].attacki, worm[i][j].attackj, other, 
	      "small_semeai_analyzer", i, j)) {
    int acode=attack(m, n, &ai, &aj);
    if (acode == 0) {
      popgo();
      change_defense(m, n, worm[i][j].attacki, worm[i][j].attackj, 1);
    }
    else if (trymove(ai, aj, color,
		     "small_semeai_analyzer", i, j)) {
      if (attack(i, j, NULL, NULL) == 0) {
	popgo();
	popgo();
	change_attack(i, j, -1, -1, 0);
      }
      else {
	popgo();
	popgo();
      }
    }
  }
  if (trymove(worm[m][n].attacki, worm[m][n].attackj, color, 
	      "small_semeai_analyzer", m, n)) {
    int acode=attack(i, j, &ai, &aj);
    if (acode == 0) {
      popgo();
      change_defense(i, j, worm[m][n].attacki, worm[m][n].attackj, 1);
    }
    else if (trymove(ai, aj, other,
		     "small_semeai_analyzer", m, n)) {
      if (attack(m, n, NULL, NULL) == 0) {
	popgo();
	popgo();
	change_attack(m, n, -1, -1, 0);
      }
      else {
	popgo();
	popgo();
      }
    }
  }
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
